unit CustomRP;

interface

uses
{$IFDEF WIN32}
  Windows, ComCtrls,
{$ELSE}
  WinTypes, WinProcs,
{$ENDIF}
  Forms, Messages, SysUtils, Classes, Controls, ExtCtrls, Dialogs,
  Graphics, TypInfo, Grids;

const
  INVALID_CONTROLCOEFICIENT = -1;
  INVALID_CONTROLSIZE = -1000;

type
  { Each TMRControlHandler handles the resizing of a single control }
  TMRControlHandler = class(TObject)
  protected
    FControl: TControl;
    FHandlers: TList;

    CLeft, CTop, CWidth, CHeight: Single;
    LastPos: TRect;
    LastParentWidth, LastParentHeight: Integer;
  public
    constructor Create( Control: TControl ); virtual;
    destructor Destroy; override;
    procedure BeforeResize( ParentWidth, ParentHeight: Integer);
    procedure AfterResize;
    function GetWarnings: string;

    procedure CalcOldPosition( ParentWidth, ParentHeight: Integer ); virtual;
    procedure CalcPropertiesOldPositions; virtual;
    function CalcNewPosition: TRect; virtual;
    procedure CalcPropertiesNewPositions; virtual;

    property Control: TControl read FControl;
    property Handlers: TList read FHandlers;
  end;
  TMRControlHandlerClass = class of TMRControlHandler;

  TMRPropertyHandler = class(TObject)
  protected
    FControl: TControl;
    procedure Change;       (* BUGBUG  Just for debugging *)

  public
    Invalid: Boolean;

    constructor Create( Control: TControl ); virtual;
    procedure BeforeResize; virtual;
    procedure AfterResize; virtual;
    function GetWarning: string; virtual;

    property Control: TControl read FControl;
  end;

  (***)

  TMRFont = class(TMRPropertyHandler)
  private
    FFont: TFont;
    FName: string;
    LastSize: Integer;
    CxSize, CySize: Single;
  public
    constructor Createhandler( Control: TControl; PropertyName: string ); virtual;
    procedure BeforeResize; override;
    procedure BeforeFontResize( CWidth, CHeight: Integer);
    procedure AfterResize; override;
    function GetWarning: string; override;

    property Font: TFont read FFont write FFont;
  end;

  TCustomResizerPanel = class(TPanel)
  private
    FPreviousWidth, FPreviousHeight: Integer;
    FControlHandlers: TList;
    FControlHandlerClass: TMRControlHandlerClass;
    FWarnings: TStringList;
    FFont: TMRFont;
    FActive: Boolean;
    FAllowChanges: Boolean;

    procedure InsertNewControls( ParentControl: TWinControl );
    procedure SetActive( Value: Boolean );

  protected
    function GetWarnings: TStringList;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure AddHandlers( Control: TControl; List: TList ); virtual;
    procedure Loaded; override;

    property Active: Boolean read FActive write SetActive;
    property ControlHandlerClass: TMRControlHandlerClass
      read FControlHandlerClass write FControlHandlerClass;

  public
    constructor Create( AOwner: TComponent ); override;
    destructor Destroy; override;
    procedure RefreshSize;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;

    property Warnings: TStringList read GetWarnings;

  published
    property AllowChanges: Boolean read FAllowChanges write FAllowChanges;
  end;


  function GetPropertyInfo( ClassInfo: Pointer; PropName: String ): Pointer;


implementation


constructor TCustomResizerPanel.Create( AOwner: TComponent );
begin
  inherited Create( AOwner );

  {$IFDEF WIN32}
  ControlStyle := [csAcceptsControls, csCaptureMouse, csClickEvents,
    csSetCaption, csOpaque, csDoubleClicks, csReplicatable];
  {$ELSE}
  ControlStyle := [csAcceptsControls, csCaptureMouse, csClickEvents,
    csSetCaption, csOpaque, csDoubleClicks];
  {$ENDIF}

  FControlHandlerClass := TMRControlHandler;

  FActive := True;

  FControlHandlers := TList.Create;
  FFont := TMRFont.CreateHandler( TControl(Self), 'Font' );
  FWarnings := TStringList.Create;

{$IFDEF DEBUG}
  if (FindWindow('TAppBuilder', nil) = 0) or
    (FindWindow('TAlignPalette', nil) = 0) or
    (FindWindow('TPropertyInspector', nil) = 0) or
    (FindWindow('TMenuBuilder', nil) = 0) or
    (FindWindow('TApplication', nil) = 0) then
  begin
    MessageBeep(1);
    MessageDlg('ResizerPanel component Not Registered!', mtWarning, [MBOK], 0);
  end;
{$ENDIF}
end;


destructor TCustomResizerPanel.Destroy;
var
  Cnt: Integer;
begin
  FFont.Free;
  FWarnings.Free;

  for Cnt := FControlHandlers.Count-1 downto 0 do
    TMRControlHandler(FControlHandlers.Items[Cnt]).Free;
  FControlHandlers.Free;
  FControlHandlers := nil;

  inherited Destroy;
end;


procedure TCustomResizerPanel.SetActive( Value: Boolean );
{ Activate/Deactive the component. When deactivated all controls
 are freed from memory, no current positions are maintained
 and no more resizing occurs. }
var
  Cnt: Integer;
begin
  if (Value <> FActive) then
  begin
    FActive := Value;

    if (Value = False) then
    begin
      for Cnt := FControlHandlers.Count-1 downto 0 do
        TMRControlHandler(FControlHandlers.Items[Cnt]).Free;

      FControlHandlers.Clear;
    end;
  end;
end;


procedure TCustomResizerPanel.Notification(AComponent: TComponent;
  Operation: TOperation);
{ This function is called whenever a child component is added or removed.
 Whenever a control is removed it's status is removed from memory.}
var
  Cnt: Integer;
begin
  inherited Notification(AComponent, Operation);

  (* check to see if it's a control *)
  if ((AComponent is TControl) and (Operation = opRemove)
     and (FControlHandlers <> nil)) then
  begin
    (* remove control from list *)
    For Cnt := FControlHandlers.Count-1 downto 0 do
      if (TMRControlHandler(FControlHandlers.Items[Cnt]).Control = AComponent)
        then
      begin
        TMRControlHandler(FControlHandlers.Items[Cnt]).Free;
        FControlHandlers.Delete( Cnt );
        break;
      end;
  end;
end;


procedure TCustomResizerPanel.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  { The WMSize is often called inside SetBounds so this gotta be here }
  if (not FActive) then
  begin
    FPreviousWidth := Width;
    FPreviousHeight := Height;
  end;

  inherited SetBounds(ALeft, ATop, AWidth, AHeight);

  if (not FActive) then
  begin
    FPreviousWidth := Width;
    FPreviousHeight := Height;
  end;
end;


procedure TCustomResizerPanel.WMSize(var Message: TWMSize);
{ A resizing occurred. Process it! }
var
  Cnt: Integer;
  Handler: TMRControlHandler;
  VState: Boolean;
begin
  { if it's in design time or just reading the components states or even
    if the component is deactivated nothing occurs and bails out right now }
  if (csDesigning in ComponentState) or (csReading in ComponentState) or
    (FActive = False) then
  begin
    inherited;
    FPreviousWidth := Width;
    FPreviousHeight := Height;
    exit;
  end;

  { insert newer controls, the ones which were inserted since the last resizing }
  InsertNewControls( self );

  VState := Visible;
  try
    (* disables drawing *)
{    LockWindowUpdate( Handle ); }
    Visible := False;

    (* Before Resize *)
    if FAllowChanges then
    begin
      FFont.BeforeFontResize( FPreviousWidth, FPreviousHeight );

      for Cnt := 0 to FControlHandlers.Count-1 do
      begin
        Handler := FControlHandlers.Items[Cnt];

        if (Handler.Control.Parent = self) then
          Handler.BeforeResize( FPreviousWidth, FPreviousHeight )
        else
          Handler.BeforeResize( Handler.Control.Parent.ClientWidth,
            Handler.Control.Parent.ClientHeight );
      end;
    end;

    inherited;

    (* After Resize *)
    FFont.AfterResize;  (* adjusts own font size *)

    for Cnt := 0 to FControlHandlers.Count-1 do
      TMRControlHandler(FControlHandlers.Items[Cnt]).AfterResize;

  finally
    (* enables drawing *)
{    LockWindowUpdate( 0 ); }
    Visible := VState;

    FPreviousWidth := Width;
    FPreviousHeight := Height;
  end;
end;


procedure TCustomResizerPanel.RefreshSize;
var
  Cnt: Integer;
  VState: Boolean;
begin
  VState := Visible;
  try
    (* disables drawing *)
    Visible := False;

    FFont.AfterResize;  (* adjusts own font size *)

    for Cnt := 0 to FControlHandlers.Count-1 do
      TMRControlHandler(FControlHandlers.Items[Cnt]).AfterResize;

  finally
    (* enables drawing *)
    Visible := VState;
  end;
end;


procedure TCustomResizerPanel.InsertNewControls( ParentControl: TWinControl );
var
  Cnt, Cnt2: Integer;
  Control: TControl;
  Exist: Boolean;
  Ch: TMRControlHandler;
begin
  with ParentControl do
  begin
    (* cycle between all child controls *)
    For Cnt := 0 to ControlCount-1 do
    begin
      Control := Controls[Cnt];

      (* sees if control already exists in control list *)
      Exist := False;
      For Cnt2 := 0 to FControlHandlers.Count-1 do
        if (TMRControlHandler(FControlHandlers.Items[Cnt2]).Control = Control) then
        begin
          Exist := True;
          break;
        end;

      if (Exist = False) then
      begin
        (* insert control handler in list *)
        Ch := FControlHandlerClass.Create( Control );
        FControlHandlers.Add( Ch );
        AddHandlers( Control, Ch.Handlers );

        if not FAllowChanges then 
          if (Ch.Control.Parent = self) then
            Ch.BeforeResize( FPreviousWidth, FPreviousHeight )
          else
            Ch.BeforeResize( Ch.Control.Parent.ClientWidth,
              Ch.Control.Parent.ClientHeight );
      end;

      if (Control is TWinControl) and not(Control is TCustomResizerPanel) then
        InsertNewControls( TWinControl(Control) );
    end;
  end;
end;


procedure TCustomResizerPanel.AddHandlers( Control: TControl; List: TList );
begin
end;


procedure TCustomResizerPanel.Loaded;
begin
  FPreviousWidth := Width;
  FPreviousHeight := Height;

  inherited Loaded;
end;


function TCustomResizerPanel.GetWarnings: TStringList;
{ Returns a list of warnings (things to avoid)
 which are calculated by 'asking' all the controls
 what seems to be wrong }
var
  Cnt: Integer;
  st: String;
  c: TWinControl;
begin
  FWarnings.Clear;

  c := Parent;
  while (c <> nil) do
  begin
    if (c is TForm) then
      if TForm(c).Scaled then
        FWarnings.AddObject( 'Turn scaled property to false', c );
    c := c.Parent;
  end;

  InsertNewControls( self );

  for Cnt := 0 to FControlHandlers.Count-1 do
  begin
    st := TMRControlHandler(FControlHandlers.Items[Cnt]).GetWarnings;
    if (st <> '') then
      FWarnings.AddObject( st,
        TMRControlHandler(FControlHandlers.Items[Cnt]).Control );
  end;

  Result := FWarnings;
end;


{*************************************************************
*************************************************************
              TMRControlHandler
*************************************************************
*************************************************************}

constructor TMRControlHandler.Create( Control: TControl );
begin
  CLeft := INVALID_CONTROLCOEFICIENT;

  FControl := Control;
  FHandlers := TList.Create;
end;


destructor TMRControlHandler.Destroy;
begin
  while (FHandlers.Count > 0) do
  begin
    TMRPropertyHandler(FHandlers.Items[0]).Destroy;
    FHandlers.Delete(0);
  end;

  FHandlers.Free;

  inherited Destroy;
end;


procedure TMRControlHandler.BeforeResize( ParentWidth, ParentHeight: Integer );
begin
  CalcOldPosition( ParentWidth, ParentHeight );

  CalcPropertiesOldPositions;
end;


procedure TMRControlHandler.AfterResize;
begin
  if (CLeft = INVALID_CONTROLCOEFICIENT) then
    exit;

  CalcNewPosition;

  CalcPropertiesNewPositions;

  (* save current position *)
  LastPos := FControl.BoundsRect;
end;


procedure TMRControlHandler.CalcOldPosition( ParentWidth, ParentHeight: Integer );
var
  NPos: TRect;
  cl, ct, cw, ch, pw, ph: Single;
begin
  (* Get control position and dimension *)
  NPos := FControl.BoundsRect;

  (* verify if any change occurred in the control size or position
   relative to the parent *)
  if ((LastParentWidth <> ParentWidth) or (LastParentHeight <> ParentHeight) or
    (LastPos.Left <> NPos.Left) or (LastPos.Right <> NPos.Right) or
    (LastPos.Top <> NPos.Top) or (LastPos.Bottom <> NPos.Bottom)) then
    if ((ParentWidth = 0) or (ParentHeight = 0)) then
    begin
      CLeft := INVALID_CONTROLCOEFICIENT;
      LastParentWidth := INVALID_CONTROLSIZE;
    end
    else
    begin
      (* calculate position coeficients *)
      cl := FControl.Left;
      ct := FControl.Top;
      cw := FControl.Width;
      ch := FControl.Height;
      pw := ParentWidth;
      ph := ParentHeight;
      CLeft := cl / pw;
      CTop := ct / ph;
      CWidth := cw / pw;
      CHeight := ch / ph;
    end;
end;


procedure TMRControlHandler.CalcPropertiesOldPositions;
var
  Cnt: Integer;
begin
  (* Treat other properties *)
  for Cnt := 0 to FHandlers.Count-1 do
    TMRPropertyHandler(FHandlers.Items[Cnt]).BeforeResize;
end;


function TMRControlHandler.CalcNewPosition: TRect;
begin
  (* Get control parent's dimensions *)
  LastParentWidth := FControl.Parent.ClientWidth;
  LastParentHeight := FControl.Parent.ClientHeight;

  (* control new dimension *)
  Result := Bounds(
    Round(CLeft*LastParentWidth),
    Round(CTop*LastParentHeight),
    Round(CWidth*LastParentWidth),
    Round(CHeight*LastParentHeight) );

  FControl.BoundsRect := Result;
end;


procedure TMRControlHandler.CalcPropertiesNewPositions;
var
  Cnt: Integer;
begin
  (* Treat other control properties *)
  for Cnt := 0 to FHandlers.Count-1 do
    TMRPropertyHandler(FHandlers.Items[Cnt]).AfterResize;
end;


function TMRControlHandler.GetWarnings: string;
var
  Cnt: Integer;
  st: string;
begin
  Result := '';

  for Cnt := 0 to FHandlers.Count-1 do
  begin
    st := TMRPropertyHandler(FHandlers.Items[Cnt]).GetWarning;
    Result := Result + st;
  end;
end;


(**************************************
**************************************
**************************************)

constructor TMRPropertyHandler.Create( Control: TControl );
begin
  FControl := Control;
  Invalid := True;
end;

procedure TMRPropertyHandler.BeforeResize;
begin
end;

procedure TMRPropertyHandler.AfterResize;
begin
end;

function TMRPropertyHandler.GetWarning: string;
begin
  Result := '';
end;

procedure TMRPropertyHandler.Change;
begin
{  MessageBeep(1); }
end;


(**************************************
**************************************
**************************************
 ******* Font property handler *******)

constructor TMRFont.CreateHandler( Control: TControl; PropertyName: string );
var
  pInfo: PPropInfo;
begin
  Create( Control );
  FName := PropertyName;

  (* get TFont instance *)
  pInfo := GetPropertyInfo( Control.ClassInfo, FName );
  FFont := TFont(GetOrdProp(Control, pInfo));
end;


procedure TMRFont.BeforeResize;
var
  fs: Single;
  W, H: Single;
begin
  (* verifies if the font size changed *)
  fs := FFont.Size;
  if ((LastSize <> Trunc(fs)) or (Invalid = True)) then
  begin
    Change; (* DEBUG *)

    W := FControl.ClientWidth;
    H := FControl.ClientHeight;
    if ((W = 0) or (H = 0)) then
    begin
      Invalid := True;
      exit;
    end;

    CxSize := fs / W;
    CySize := fs / H;
  end;

  Invalid := False;
end;

procedure TMRFont.BeforeFontResize( CWidth, CHeight: Integer);
(* Aditional function just for handling the MultiResPanel font
 which simple replaces the current dimensions with the given *)
var
  fs: Single;
  W, H: Single;
begin
  (* verifies if the font size changed *)
  fs := FFont.Size;
  if ((LastSize <> Trunc(fs)) or (Invalid = True)) then
  begin
    Change; (* DEBUG *)

    W := CWidth;
    H := CHeight;
    if ((W = 0) or (H = 0)) then
    begin
      Invalid := True;
      exit;
    end;

    CxSize := fs / W;
    CySize := fs / H;
  end;

  Invalid := False;
end;


procedure TMRFont.AfterResize;
var
  SizeX, SizeY: Single;
  H, W: Single;
begin
  if (Invalid = True) then
    exit;

  W := FControl.ClientWidth;
  H := FControl.ClientHeight;
  SizeX := CxSize * W;
  SizeY := CySize * H;

  if (SizeX < SizeY) then
    FFont.Size := Trunc(SizeX)
  else
    FFont.Size := Trunc(SizeY);

  if (FFont.Size = 0) then
    FFont.Size := 1;

  LastSize := FFont.Size;

  {$IFDEF WIN32}
  if ((FControl is TTabSheet) and (FControl.Parent is TPageControl)) then
    FControl.BoundsRect := TPageControl(FControl.Parent).ActivePage.BoundsRect;
  {$ENDIF}
end;


function TMRFont.GetWarning: string;
var
  tm: TTextMetric;
  OldF, hDC: THandle;
begin
  hDC := GetDC( 0 );
  try
    OldF := SelectObject( hDC, FFont.Handle );
    GetTextMetrics( hDC, tm );
    SelectObject( hDC, OldF );
  finally
    ReleaseDC( 0, hDC );
  end;

  if ((tm.tmPitchAndFamily and TMPF_VECTOR) = 0) then
    Result := 'Not a vector font. '
  else
    Result := '';
end;


function GetPropertyInfo( ClassInfo: Pointer; PropName: String ): Pointer;
begin
  Result := GetPropInfo( ClassInfo, PropName );
  if (Result = nil) then
    raise Exception.Create('No ' + PropName + ' property');
end;

end.
